"""Class: Forms

Posts requests to forms/ endpoints. Requires a valid LazarusAuth
instance on initialization. Including the optional model_id argument
will create a custom Forms instance, using the model corresponding to
the model_id over the generic Forms model.
"""

import sys
import os
import requests

from .lazarus_auth import LazarusAuth

parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(parent_dir)

import utils

BASE_URL = os.environ.get("BASE_URL", "https://api.lazarusforms.com/")


class Forms:
    """A class to post requests to all forms/ endpoints."""

    def __init__(self, auth: LazarusAuth, model_id=None):
        """Initialize a Forms() object.

        Without model_id, creates a Forms() object that uses the generic
        forms model. With model_id, creates a Forms() object that uses
        the custom model corresponding to the model_id.

        Args:
            auth (LazarusAuth): Holds authenticated header information
            model_id (str, optional): Custom model ID, defaults to None
        """
        self.headers = auth.headers
        self.model_id = model_id


    def run_ocr(self, input_type, input_str, **kwargs):
        """Posts a request to the relevant forms/ endpoint.

        If the Forms instance was not initialized with a model_id, we post to
        the api/forms/generic endpoint. If a model_id was supplied on init, we
        post to the api/forms/custom/{model_id} endpoint.

        Args:
            input_type (str): Type of input expected [FILE_PATH, URL, BASE64]
            input_str (str): File to upload, expecting a file path, url, or a base64 encoded string
            kwargs (dict, optional): Must include at least one of the following fields
                file_id (str): Custom ID for the uploaded document
                metadata (dict): Data to be returned in the response
                webhook (str): Webhook to ping after call to API
        """
        possible_kwargs = ["file_id", "metadata", "webhook"]
        kwargs = utils._validate_args(kwargs, possible_kwargs)

        if self.model_id is not None:
            url = f"{BASE_URL}/api/forms/custom/{self.model_id}"
        else:
            url = f"{BASE_URL}/api/forms/generic"

        headers = self.headers | utils._get_typed_headers(input_type)
        data = utils._get_typed_body(input_type, input_str)

        if input_type == "FILE_PATH":
            response = requests.post(url, headers=headers, files=data, data=kwargs)
        else:
            response = requests.post(url, headers=headers, json=data | kwargs)

        if response.ok:
            resp = response.json()
            utils._record_metrics("forms", self.headers, self.model_id, resp)
            return resp

        utils._record_metrics("forms", self.headers, self.model_id)
        utils._error_handling(response)


# Forms class usage examples
if __name__ == "__main__":
    # Create a LazarusAuth object with your org ID and auth key
    org_id = os.environ.get("LAZARUS_ORG_ID")
    auth_key = os.environ.get("LAZARUS_AUTH_KEY")
    auth = LazarusAuth(org_id, auth_key)

    # Create a Forms object
    forms = Forms(auth)
    # Create a custom Forms object, model_id is your custom Forms model
    custom_forms = Forms(auth, model_id="MODEL_ID_HERE")

    # Upload a file using a path to a file
    response = forms.run_ocr("FILE_PATH", "/path/to/file.pdf")
    # Upload a file using a URL
    response = forms.run_ocr("URL", "https://fileurl.com")
    # Upload a file using a Base64 encoded string
    response = forms.run_ocr("BASE64", "base64_encoded_string")

    # Upload a file using optional args
    kwargs = {
        "file_id": "filename", 
        "metadata": {"foo": "bar"}, 
        "webhook": "https://pingme.com"
    }
    forms.run_ocr("URL", "https://fileurl.com", **kwargs)
    forms.run_ocr("URL", "https://fileurl.com", file_id="filename", metadata={"foo": "bar"}, webhook="https://pingme.com")



from .args_validation import _validate_args
from .error_handling import _error_handling
from .input_types import _get_typed_headers, _get_typed_body, _get_multipart_data
from .metrics import _record_metrics



"""Class: RikAI

Posts requests to api/rikai/ endpoints. Requires a valid LazarusAuth
instance on initialization. Including the optional model_id argument
will create a custom RikAI instance, using the model corresponding to
the model_id over the standard RikAI model.
"""

import os
import sys
import requests

from .lazarus_auth import LazarusAuth

parent_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.append(parent_dir)

import utils

BASE_URL = os.environ.get("BASE_URL", "https://api.lazarusforms.com/")


class RikAI:
    """A class to post requests to all rikai/ endpoints."""

    def __init__(self, auth: LazarusAuth, model_id=None):
        """Initialize a RikAI() object.

        Without model_id, creates a RikAI() object that uses the standard
        RikAI model. With model_id, creates a RikAI() object that uses
        the custom model corresponding to the model_id.

        Args:
            auth (LazarusAuth): Holds authenticated header information
            model_id (str, optional): Custom model ID, defaults to None
        """
        self.headers = auth.headers
        self.model_id = model_id


    def ask_question(self, input_type: str, input_str: str, question: list, **kwargs):
        """Posts a request to the relevant rikai/ endpoint.

        If the RikAI instance was not initialized with a model_id, we post to
        the api/rikai endpoint. If a model_id was supplied on init, we post
        to the api/rikai/custom/{model_id} endpoint.

        Args:
            input_type (str): Type of input expected [FILE_PATH, URL, BASE64]
            input_str (str): File to upload, expecting a file path, url, or a base64 encoded string
            question (list): A list of strings containing the question(s) to be asked
            kwargs (dict, optional): Must include at least one of the following fields
                file_id (str): Custom ID for the uploaded document
                metadata (dict): Data to be returned in the response
                webhook (str): Webhook to ping after call to API
                settings (dict): User settings specified in the request, for custom RikAI only
                return_ocr (bool): Set to True to add OCR results to the response, defaults to False
                language (str): A 2 character language code or the name of the language you wish to translate answers into
        """
        url = f"{BASE_URL}/api/rikai"
        possible_kwargs = ["file_id", "metadata", "webhook", "return_ocr", "language"]
        if self.model_id is not None:
            url += f"/custom/{self.model_id}"
            possible_kwargs.append("settings")

        kwargs = utils._validate_args(kwargs, possible_kwargs)

        headers = self.headers | utils._get_typed_headers(input_type)
        body = utils._get_typed_body(input_type, input_str)

        if input_type == "FILE_PATH":
            data = {"question": question} | kwargs
            response = requests.post(url, headers=headers, files=body, data=data)
        else:
            body |= {"question": question} | kwargs
            response = requests.post(url, headers=headers, json=body)

        if response.ok:
            resp = response.json()
            utils._record_metrics("rikai", self.headers, self.model_id, resp)
            return resp

        utils._record_metrics("rikai", self.headers, self.model_id)
        utils._error_handling(response)


    def summarize(self, input_type: str, input_str: str, fields: dict):
        """Posts a request to the rikai/summarize endpoint.

        Args:
            input_type (str): Type of input expected [URL, BASE64]
            input_str (str): File to upload, expecting a url or a base64 encoded string
            fields (dict): Required fields to prompt the summarizer
                document_type (str): Type of document to summarize
                summary_description (str): Description of what information should be included in the summary
                secondary_description (str, optional): A secondary summary description, including this will return a secondary summary
                json_format (str, optional): Specify a JSON output structure, content will be pulled from the resulting summary description
        """
        if input_type == "FILE_PATH":
            raise ValueError("Summarize only accepts \"URL\" and \"BASE64\" input types")

        url = f"{BASE_URL}/api/rikai/summarize"
        fields = utils._validate_args(fields, ["secondary_description", "json_format"], ["document_type", "summary_description"])

        headers = self.headers | utils._get_typed_headers(input_type)
        body = utils._get_typed_body(input_type, input_str) | {"fields": fields}

        response = requests.post(url, headers=headers, json=body)

        if response.ok:
            resp = response.json()
            utils._record_metrics("rikai/summarizer", self.headers, self.model_id, resp)
            return resp

        utils._record_metrics("rikai/summarizer", self.headers, self.model_id)
        utils._error_handling(response)


# RikAI class usage examples
if __name__ == "__main__":
    # Create a LazarusAuth object with your org ID and auth key
    org_id = os.environ.get("LAZARUS_ORG_ID")
    auth_key = os.environ.get("LAZARUS_AUTH_KEY")
    auth = LazarusAuth(org_id, auth_key)

    # Create a RikAI object
    rikai = RikAI(auth)
    # Create a custom RikAI object, model_id_here is your custom RikAI model ID
    custom_rikai = RikAI(auth, "model_id_here")

    # RikAI Q&A
    questions = ["What is this document about?", "When was this document published?"]
    # Upload a file using a local file path
    response = rikai.ask_question("FILE_PATH", "/path/to/file.pdf", questions)
    # Upload a file using a URL
    response = rikai.ask_question("URL", "https://fileurl.com", questions)
    # Upload a file using a Base64 encoded string
    response = rikai.ask_question("BASE64", "base64_encoded_string", questions)

    # RikAI Summarize
    fields = {
        "document_type": "A medical form",
        "summary_description": "List all patient personal information such as DOB and address",
        "secondary_description": "Any information related to the health of the patient",
        "json_format": "{\"Patient Name\":\"John Smith\",\"Patient Address\":\"1234 Home Drive, Boston, MA 02112\"}"
    }
    # Upload a file using a URL
    response = rikai.summarize("URL", "https://fileurl.com", fields)
    # Upload a file using a Base64 encoded string
    response = rikai.summarize("BASE64", "base64_encoded_string", fields)

    # Upload a file using optional args
    kwargs = {
        "return_ocr": True,
        "language": "Japanese",
        "settings": {
            "verbose": True,
        },
        "file_id": "filename",
        "metadata": {"foo": "bar"},
        "webhook": "https://pingme.com"
    }
    rikai.ask_question("URL", "https://fileurl.com", questions, **kwargs)
    rikai.ask_question("URL", "https://fileurl.com", questions, return_ocr=True, language="Japanese")



"""Helper function to convert responses from failed requests to errors."""

import json
from requests import Response

from errors import APIError, AuthError


def _error_handling(response: Response):
    """Processes response with non-200 response code.

    If the response data is JSON serializable, use the data from the response
    in the error message. Otherwise, raises an error with just the status and
    error code.

    Args:
        response (Response): Response from post request
    Raises:
        APIError if response was not successful
        AuthError if org_id or auth_key were invalid
    """
    try:
        resp = response.json()

        # If there's more than just status and message in response,
        # include the response in api output
        api_output = None
        if resp.keys() > {"status", "message"}:
            api_output = resp

        if response.status_code == 403:
            raise AuthError("AUTH_FAILURE", resp["message"], 403, api_output)
        raise APIError("FAILURE", resp["message"], response.status_code, api_output)
    except (json.JSONDecodeError, KeyError):
        pass

    # If the response is not a json, return static info
    if response.status_code == 403:
        raise AuthError("AUTH_FAILURE", "Invalid authentication.", 403)
    raise APIError("FAILURE", None, response.status_code)



"""Helper functions used to map input types to requests and responses."""

import os
import sys

parent_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(parent_dir)

# Mapping file extensions to their MIME types
FILE_EXTENSIONS = {
    '.pdf': 'application/pdf',
    '.png': 'image/png',
    '.jpeg': 'image/jpg',
    '.jpg': 'image/jpg',
    '.tif': 'image/tiff',
    '.tiff': 'image/tiff',
    '.webp': 'image/webp'
}


def _get_typed_headers(input_type: str) -> dict:
    """Maps input_type to request headers.

    Args:
        input_type (str): Type of input expected [FILE_PATH, URL, BASE64]
    Returns:
        dict: Content type key-value pair
    """
    match input_type:
        case "FILE_PATH":
            return {}
        case "URL" | "BASE64":
            return {"Content-Type": "application/json"}
        case _:
            raise ValueError("Expected one of: \"FILE_PATH\", \"URL\", \"BASE64\"")


def _get_typed_body(input_type: str, input_str: str):
    """Maps input_type and input_str to request body fields.

    Args:
        input_type (str): Type of input expected [FILE_PATH, URL, BASE64]
        input_str (str): A path to a file, url or a base64 encoded string
    Returns:
        dict: Request body key-value pairs
    """
    match input_type:
        case "FILE_PATH":
            return {"file": _get_multipart_data(input_str)}
        case "URL":
            return {"inputUrl": input_str}
        case "BASE64":
            return {"base64": input_str}
        case _:
            raise ValueError("Expected one of: \"FILE_PATH\", \"URL\", \"BASE64\"")


def _get_multipart_data(path: str) -> tuple:
    """Converts a file path into Multipart Encoded data.

    Extracts necessary info from the path, extends it to an absolute
    path, and creates a tuple holding file information that can be passed
    to a request in the data parameter. This function will raise a 
    FileNotFoundError if the path supplied is invalid.

    Args:
        path (str): File path
    Returns:
        tuple: Corresponds to (file name, read file object, MIME type)
    """
    filename = os.path.basename(path)
    _, ext = os.path.splitext(filename)
    path = os.path.join(os.path.abspath(""), path)

    if ext not in FILE_EXTENSIONS:
        raise ValueError(f"File must be one of: {FILE_EXTENSIONS.keys()}")

    return (filename, open(path, 'rb'), FILE_EXTENSIONS[ext])



"""Helper function to validate user-provided args."""
import stringcase

from errors import ValidationError


def _validate_args(args: dict, valid_keys: list, ignore_keys: list = [], change_case: bool = True) -> dict:
    """Validates arguments passed by the user.

    Checks that args includes only valid or optional keys. Expecting keys in
    snake_case. Converts arguments to camelCase by default as the API expects
    keys in camelCase. Raises an error if args has undefined keys or keys not
    in snake_case.

    Args:
        args (dict): Arguments passed to library functions
        valid_keys (list): List of valid optional arguments
        ignore_keys (list, optional): List of arguments in args that don't
            need to be validated, defaults to []
        change_case (bool, optional): Boolean which is true if want to make
            the args in dictionary camel case, defaults to True
    """
    invalid_keys = []
    for key in args:
        if key not in valid_keys and key not in ignore_keys:
            invalid_keys.append(key)

    if invalid_keys:
        m = f"These fields are not valid arguments: {', '.join(invalid_keys)}"
        raise ValidationError(m)

    if change_case:
        args = _make_camel_case(args)
    return args


def _make_camel_case(args: dict) -> dict:
    """ Make args camel case for API request body

    Args:
        args (dict): Dictionary with keys that are mix of snake case and
            already camel case

    Returns:
        dict: Dictionary with only camel case keys and the same corresponding
            values as passed in
    """
    d = {}
    for key in args:
        if key == "return_ocr":
            d["returnOCR"] = args[key]
            continue
        new_key = stringcase.camelcase(key)
        d[new_key] = args[key]
    return d



""" Lazarus Forms Library Errors

Common errors that arise in using the library. Designed to prevent
hard to trace API Errors.
"""


class ValidationError(Exception):
    """Raised when Library encounters validation error"""


class InvalidAuthError(Exception):
    """Raised when initializing LazarusAuth encounters invalid auth info"""
